
#include <d3d9.h>
#include "output.h"

void DumpAdapter( Output& output, IDirect3D9* p_d3d, UINT adapter );
void ClearMem( void* ptr, size_t size );
const char* GetFormatString( D3DFORMAT format );

extern D3DFORMAT formats9[];
extern int num_formats9;

void DumpD3DCAPS9( Output& output )
{
    // Load Direct3D9 library
    HMODULE d3d9dll = LoadLibrary( "D3D9.DLL" );
    if ( d3d9dll == NULL ) {
	output.Error( "Cannot load D3D9.DLL" );
	return;
    }

    // Get access to Direct3DCreate9 function
    typedef IDirect3D9* (WINAPI *DIRECT3DCREATE9)( UINT );
    DIRECT3DCREATE9 procDirect3DCreate9 = (DIRECT3DCREATE9)GetProcAddress( d3d9dll, "Direct3DCreate9" );
    if ( procDirect3DCreate9 == NULL ) {
	output.Error( "Cannot obtain pointer to Direct3DCreate9" );
	FreeLibrary( d3d9dll );
	return;
    }

    // Get access to IDirect3D9
    const int sdk9_version = 31;
    IDirect3D9* p_d3d = procDirect3DCreate9( sdk9_version );
    if ( p_d3d == NULL ) {
	output.Error( "Cannot obtain IDirect3D9 interface" );
	FreeLibrary( d3d9dll );
	return;
    }

    // Dump caps of subsequent adapters
    if ( p_d3d->GetAdapterCount() < 2 ) {
	DumpAdapter( output, p_d3d, D3DADAPTER_DEFAULT );
    }
    else {
	for ( UINT i=0; i < p_d3d->GetAdapterCount(); ++i ) {
	    DumpAdapter( output, p_d3d, i );
	}
    }
    p_d3d->Release();

    // Unload library
    FreeLibrary( d3d9dll );
}

void DumpAdapter( Output& output, IDirect3D9* p_d3d, UINT adapter )
{
    output.Write( "D3DVERSION", 9 );
    output.Write( "ADAPTER_INDEX", adapter );

    // Dump adapter identifier
    D3DADAPTER_IDENTIFIER9 identifier;
    int error = p_d3d->GetAdapterIdentifier( adapter, D3DENUM_WHQL_LEVEL, &identifier );
    if ( ! output.CheckError( "IDirect3D9::GetAdapterIdentifier", error ) ) {
	return;
    }
    {
	SECTION( "ADAPTER_IDENTIFIER" );
	output.Write( "DRIVER", identifier.Driver );
	output.Write( "DESCRIPTION", identifier.Description );
	char ver_s[ 64 ];
	const LARGE_INTEGER& ver = identifier.DriverVersion;
	wsprintf( ver_s, "%i.%i.%i.%i", HIWORD(ver.HighPart), LOWORD(ver.HighPart), HIWORD(ver.LowPart), LOWORD(ver.LowPart) );
	output.Write( "DRIVER_VERSION", ver_s );
	output.Write( "VENDOR_ID", identifier.VendorId );
	output.Write( "DEVICE_ID", identifier.DeviceId );
	output.Write( "SUBSYS_ID", identifier.SubSysId );
	output.Write( "REVISION", identifier.Revision );
	char guid_s[ 64 ];
	const GUID& g = identifier.DeviceIdentifier;
	wsprintf( guid_s, "%08X-%04X-%04X-%02X%02X%02X%02X%02X%02X%02X%02X", g.Data1, g.Data2, g.Data3, g.Data4[0], g.Data4[1], g.Data4[2], g.Data4[3], g.Data4[4], g.Data4[5], g.Data4[6], g.Data4[7] );
	output.Write( "DEVICE_IDENTIFIER", guid_s );
	output.Write( "WHQL_LEVEL", identifier.WHQLLevel );
    }

    // Dump caps
    D3DCAPS9 caps;
    ClearMem( &caps, sizeof(caps) );
    caps.DeviceType = D3DDEVTYPE_HAL;
    caps.AdapterOrdinal = adapter;
    error = p_d3d->GetDeviceCaps( adapter, D3DDEVTYPE_HAL, &caps );
    if ( ! output.CheckError( "IDirect3D9::GetDeviceCaps", error ) ) {
	return;
    }
    {
	SECTION( "DEVCAPS" );
	output.Write( "CAPS", caps.Caps );
	output.Write( "CAPS2", caps.Caps2 );
	output.Write( "CAPS3", caps.Caps3 );
	output.Write( "PRESENTATION_INTERVALS", caps.PresentationIntervals );
	output.Write( "CURSOR_CAPS", caps.CursorCaps );
	output.Write( "DEV_CAPS", caps.DevCaps );
	output.Write( "PRIMITIVE_MISC_CAPS", caps.PrimitiveMiscCaps );
	output.Write( "RASTER_CAPS", caps.RasterCaps );
	output.Write( "ZCMP_CAPS", caps.ZCmpCaps );
	output.Write( "SRC_BLEND_CAPS", caps.SrcBlendCaps );
	output.Write( "DEST_BLEND_CAPS", caps.DestBlendCaps );
	output.Write( "ALPHA_CMP_CAPS", caps.AlphaCmpCaps );
	output.Write( "SHADE_CAPS", caps.ShadeCaps );
	output.Write( "TEXTURE_CAPS", caps.TextureCaps );
	output.Write( "TEXTURE_FILTER_CAPS", caps.TextureFilterCaps );
	output.Write( "CUBE_TEXTURE_FILTER_CAPS", caps.CubeTextureFilterCaps );
	output.Write( "VOLUME_TEXTURE_FILTER_CAPS", caps.VolumeTextureFilterCaps );
	output.Write( "TEXTURE_ADDRESS_CAPS", caps.TextureAddressCaps );
	output.Write( "VOLUME_TEXTURE_ADDRESS_CAPS", caps.VolumeTextureAddressCaps );
	output.Write( "LINE_CAPS", caps.LineCaps );
	output.Write( "MAX_TEXTURE_WIDTH", caps.MaxTextureWidth );
	output.Write( "MAX_TEXTURE_HEIGHT", caps.MaxTextureHeight );
	output.Write( "MAX_VOLUME_EXTENT", caps.MaxVolumeExtent );
	output.Write( "MAX_TEXTURE_REPEAT", caps.MaxTextureRepeat );
	output.Write( "MAX_TEXTURE_ASPECT_RATIO", caps.MaxTextureAspectRatio );
	output.Write( "MAX_ANISOTROPY", caps.MaxAnisotropy );
	output.Write( "MAX_VERTEX_W", *(DWORD*)&caps.MaxVertexW );
	output.Write( "GUARD_BAND_LEFT", *(DWORD*)&caps.GuardBandLeft );
	output.Write( "GUARD_BAND_TOP", *(DWORD*)&caps.GuardBandTop );
	output.Write( "GUARD_BAND_RIGHT", *(DWORD*)&caps.GuardBandRight );
	output.Write( "GUARD_BAND_BOTTOM", *(DWORD*)&caps.GuardBandBottom );
	output.Write( "EXTENTS_ADJUST", *(DWORD*)&caps.ExtentsAdjust );
	output.Write( "STENCIL_CAPS", caps.StencilCaps );
	output.Write( "FVF_CAPS", caps.FVFCaps );
	output.Write( "TEXTURE_OP_CAPS", caps.TextureOpCaps );
	output.Write( "MAX_TEXTURE_BLEND_STAGES", caps.MaxTextureBlendStages );
	output.Write( "MAX_SIMULTANEOUS_TEXTURES", caps.MaxSimultaneousTextures );
	output.Write( "VERTEX_PROCESSING_CAPS", caps.VertexProcessingCaps );
	output.Write( "MAX_ACTIVE_LIGHTS", caps.MaxActiveLights );
	output.Write( "MAX_USER_CLIP_PLANES", caps.MaxUserClipPlanes );
	output.Write( "MAX_VERTEX_BLEND_MATRICES", caps.MaxVertexBlendMatrices );
	output.Write( "MAX_VERTEX_BLEND_MATRIX_INDEX", caps.MaxVertexBlendMatrixIndex );
	output.Write( "MAX_POINT_SIZE", *(DWORD*)&caps.MaxPointSize );
	output.Write( "MAX_PRIMITIVE_COUNT", caps.MaxPrimitiveCount );
	output.Write( "MAX_VERTEX_INDEX", caps.MaxVertexIndex );
	output.Write( "MAX_STREAMS", caps.MaxStreams );
	output.Write( "MAX_STREAM_STRIDE", caps.MaxStreamStride );
	output.Write( "VERTEX_SHADER_VERSION", caps.VertexShaderVersion );
	output.Write( "MAX_VERTEX_SHADER_CONST", caps.MaxVertexShaderConst );
	output.Write( "PIXEL_SHADER_VERSION", caps.PixelShaderVersion );
	output.Write( "MAX_PIXEL_SHADER_1X_VALUE", *(DWORD*)&caps.PixelShader1xMaxValue );
	output.Write( "DEVCAPS2", caps.DevCaps2 );
	output.Write( "MAX_NPATCH_TESSELATION_LEVEL", *(DWORD*)&caps.MaxNpatchTessellationLevel );
	output.Write( "MASTER_ADAPTER_ORDINAL", (DWORD)caps.MasterAdapterOrdinal );
	output.Write( "ADAPTER_ORDINAL_IN_GROUP", (DWORD)caps.AdapterOrdinalInGroup );
	output.Write( "NUMBER_OF_ADAPTERS_IN_GROUP", (DWORD)caps.NumberOfAdaptersInGroup );
	output.Write( "DECL_TYPES", caps.DeclTypes );
	output.Write( "NUM_SIMULTANEOUS_RTS", caps.NumSimultaneousRTs );
	output.Write( "STRETCH_RECT_FILTER_CAPS", caps.StretchRectFilterCaps );
	output.Write( "VS20_CAPS", caps.VS20Caps.Caps );
	output.Write( "VS20_CAPS_DYNAMIC_FLOW_CONTROL_DEPTH", (DWORD)caps.VS20Caps.DynamicFlowControlDepth );
	output.Write( "VS20_CAPS_NUM_TEMPS", (DWORD)caps.VS20Caps.NumTemps );
	output.Write( "VS20_CAPS_STATIC_FLOW_CONTROL_DEPTH", (DWORD)caps.VS20Caps.StaticFlowControlDepth );
	output.Write( "PS20_CAPS", caps.PS20Caps.Caps );
	output.Write( "PS20_CAPS_DYNAMIC_FLOW_CONTROL_DEPTH", (DWORD)caps.PS20Caps.DynamicFlowControlDepth );
	output.Write( "PS20_CAPS_NUM_TEMPS", (DWORD)caps.PS20Caps.NumTemps );
	output.Write( "PS20_CAPS_STATIC_FLOW_CONTROL_DEPTH", (DWORD)caps.PS20Caps.StaticFlowControlDepth );
	output.Write( "VERTEX_TEXTURE_FILTER_CAPS", caps.VertexTextureFilterCaps );
	output.Write( "MAX_VSHADER_INSTRUCTIONS_EXECUTED", caps.MaxVShaderInstructionsExecuted );
	output.Write( "MAX_PSHADER_INSTRUCTIONS_EXECUTED", caps.MaxPShaderInstructionsExecuted );
	output.Write( "MAX_VERTEX_SHADER30_INSTRUCTION_SLOTS", caps.MaxVertexShader30InstructionSlots );
	output.Write( "MAX_PIXEL_SHADER30_INSTRUCTION_SLOTS", caps.MaxPixelShader30InstructionSlots );
    }

    // Auxiliary structures
    D3DFORMAT adapter_formats[] = { D3DFMT_X8R8G8B8, D3DFMT_A8R8G8B8, D3DFMT_A2R10G10B10, D3DFMT_X1R5G5B5, D3DFMT_A1R5G5B5, D3DFMT_R5G6B5 };
    const int num_modes = sizeof(adapter_formats) / sizeof(D3DFORMAT);
    const int num_resources = 5;
    const char* resource_names[] = { "SURFACE", "VOLUME", "TEXTURE", "VOLUMETEXTURE", "CUBETEXTURE" };
    const int num_usages = 6;
    const char* usage_names[] = { "FULLSCREEN_BACKBUFFER", "WINDOWED_BACKBUFFER", "PLAIN", "RENDERTARGET", "DEPTHSTENCIL", "DMAP" };
    DWORD usages[] = { D3DUSAGE_RENDERTARGET, D3DUSAGE_RENDERTARGET, 0, D3DUSAGE_RENDERTARGET, D3DUSAGE_DEPTHSTENCIL, D3DUSAGE_DMAP };

    // Dump subsequent adapter modes
    {
	SECTION( "ADAPTER_MODES" );
	for ( int i_mode=0; i_mode < num_modes; ++i_mode ) {
	    if ( p_d3d->GetAdapterModeCount( adapter, adapter_formats[i_mode] ) == 0 ) {
		continue;
	    }
	    SECTION( GetFormatString(adapter_formats[i_mode]) );
	    for ( int i_resource=0; i_resource < num_resources; ++i_resource ) {
		SECTION( resource_names[i_resource] );
		for ( int i_format=0; i_format < num_formats9; ++i_format ) {
		    const char* format = GetFormatString( formats9[i_format] );
		    for ( int i_usage=0; i_usage < num_usages; ++i_usage ) {
			const bool b_format_ok = SUCCEEDED(
			    p_d3d->CheckDeviceFormat(
				adapter,
				D3DDEVTYPE_HAL,
				adapter_formats[i_mode],
				usages[i_usage],
				(D3DRESOURCETYPE)(i_resource+1),
				formats9[i_format] ) );
			const bool b_device_type_ok = SUCCEEDED(
			    p_d3d->CheckDeviceType(
				adapter,
				D3DDEVTYPE_HAL,
				adapter_formats[i_mode],
				formats9[i_format],
				i_usage == 1 ) );
			if ( ((i_usage > 1) && b_format_ok) ||
			     ((i_usage < 2) && (i_resource == 0) && b_device_type_ok) ) {
			    output.Write( usage_names[i_usage], format );
			}
		    }
		}
	    }
	}
    }
}
