
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "output.h"
#include <gl/gl.h>

namespace {

    LRESULT CALLBACK WndProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
    {
	if ( uMsg == WM_CREATE ) {
	    return 0;
	}
	return DefWindowProc( hwnd, uMsg, wParam, lParam );
    }

}

void DumpExtensions( Output& output, const char* type, const char* extensions );
void ClearMem( void* ptr, size_t size );

void DumpGLExtensions( Output& output )
{
    // Load OpenGL32 library
    HMODULE gldll = LoadLibrary( "OPENGL32.DLL" );
    if ( gldll == NULL ) {
	output.Error( "Cannot load OPENGL32.DLL" );
	return;
    }

    // Get addresses of GL functions
    typedef const GLubyte* (WINAPI *GLGETSTRING)( GLenum );
    typedef HGLRC (WINAPI *WGLCREATECONTEXT)( HDC );
    typedef BOOL (WINAPI *WGLMAKECURRENT)( HDC, HGLRC );
    typedef BOOL (WINAPI *WGLDELETECONTEXT)( HGLRC );
    typedef PROC (WINAPI *WGLGETPROCADDRESS)( LPCSTR );
    GLGETSTRING proc_glGetString = (GLGETSTRING)GetProcAddress( gldll, "glGetString" );
    WGLCREATECONTEXT proc_wglCreateContext = (WGLCREATECONTEXT)GetProcAddress( gldll, "wglCreateContext" );
    WGLMAKECURRENT proc_wglMakeCurrent = (WGLMAKECURRENT)GetProcAddress( gldll, "wglMakeCurrent" );
    WGLDELETECONTEXT proc_wglDeleteContext = (WGLDELETECONTEXT)GetProcAddress( gldll, "wglDeleteContext" );
    WGLGETPROCADDRESS proc_wglGetProcAddress = (WGLGETPROCADDRESS)GetProcAddress( gldll, "wglGetProcAddress" );
    if ( (proc_glGetString == NULL) || (proc_wglCreateContext == NULL) || (proc_wglMakeCurrent == NULL) || (proc_wglDeleteContext == NULL) ) {
	output.Error( "Cannot obtain address of one of the required OpenGL functions" );
	FreeLibrary( gldll );
	return;
    }

    // Create temporary window
    const HINSTANCE module_handle = GetModuleHandle( NULL );
    const char* const classname = "temp";
    const WNDCLASS wc = { 0, WndProc, 0, 0, module_handle, 0, 0, 0, 0, classname };
    RegisterClass( &wc );
    const HWND hwnd = CreateWindowEx( 0, classname, "", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, 0, 0, module_handle, 0 );
    const HDC hdc = GetDC( hwnd );

    // Set pixel format
    PIXELFORMATDESCRIPTOR pfd;
    ClearMem( &pfd, sizeof(pfd) );
    pfd.nSize		= sizeof(pfd);
    pfd.nVersion	= 1;
    pfd.dwFlags		= PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType	= PFD_TYPE_RGBA;
    pfd.cColorBits	= 24;
    pfd.cDepthBits	= 24;
    pfd.iLayerType	= PFD_MAIN_PLANE;
    const int i_format = ChoosePixelFormat( hdc, &pfd );
    SetPixelFormat( hdc, i_format, &pfd );

    // Create OpenGL rendering context
    const HGLRC hgl = proc_wglCreateContext( hdc );
    const bool b_ok = proc_wglMakeCurrent( hdc, hgl ) != FALSE;

    // Report error
    if ( ! b_ok ) {
	if ( hgl ) {
	    output.Error( "Cannot switch to OpenGL context" );
	    proc_wglDeleteContext( hgl );
	}
	else {
	    output.Error( "Cannot create OpenGL context" );
	}
	ReleaseDC( hwnd, hdc );
	DestroyWindow( hwnd );
	FreeLibrary( gldll );
	return;
    }

    // Dump header
    output.Write( "D3DVERSION", "OPENGL" );
    {
	SECTION( "GLINFO" );
	output.Write( "VENDOR", reinterpret_cast<const char*>( proc_glGetString( GL_VENDOR ) ) );
	output.Write( "RENDERER", reinterpret_cast<const char*>( proc_glGetString( GL_RENDERER ) ) );
	output.Write( "VERSION", reinterpret_cast<const char*>( proc_glGetString( GL_VERSION ) ) );
    }

    // Dump extensions
    {
	SECTION( "EXTENSIONS" );
	const char* extensions = reinterpret_cast<const char*>( proc_glGetString( GL_EXTENSIONS ) );
	if ( extensions != NULL ) {
	    DumpExtensions( output, "GL", extensions );
	}

	if ( proc_wglGetProcAddress != NULL ) {
	    typedef const char* (WINAPI *EXTSTRINGPROC)( HDC hdc );
	    EXTSTRINGPROC wglGetExtensionsString =
		(EXTSTRINGPROC)proc_wglGetProcAddress( "wglGetExtensionsStringARB" );
	    extensions = (wglGetExtensionsString != NULL) ? wglGetExtensionsString( hdc ) : NULL;
	    if ( extensions != NULL ) {
		DumpExtensions( output, "WGL", extensions );
	    }
	}
    }

    // Release contexts and window
    proc_wglMakeCurrent( 0, 0 );
    if ( hgl ) proc_wglDeleteContext( hgl );
    ReleaseDC( hwnd, hdc );
    DestroyWindow( hwnd );

    // Unload library
    FreeLibrary( gldll );
}

void DumpExtensions( Output& output, const char* type, const char* extensions )
{
    static char buf[256];
    while ( *extensions ) {
	int i = 0;
	while ( *extensions && (*extensions != ' ') ) {
	    buf[i++] = *(extensions++);
	}
	if ( i > 0 ) {
	    buf[i] = 0;
	    output.Write( type, buf );
	}
	if ( *extensions ) {
	    ++extensions;
	}
    }
}