
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "output.h"

namespace {

    size_t StrLen( const char* s )
    {
	size_t i = 0;
	while ( *(s++) ) ++i;
	return i;
    }

}

Output::Output( const char* filename )
: file( CreateFile( filename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, NULL ) )
, indent( 0 )
{
}

Output::~Output()
{
    if ( file != INVALID_HANDLE_VALUE ) {
	CloseHandle( file );
	file = INVALID_HANDLE_VALUE;
    }
}

bool Output::WriteRaw( const char* s, unsigned size )
{
    DWORD num_writ = 0;
    return
	(file != INVALID_HANDLE_VALUE) &&
	WriteFile( file, s, size, &num_writ, NULL ) &&
	(num_writ == size);
}

bool Output::Write( const char* s1, const char* s2 )
{
    // Write indentation
    static const char tabs[] = "\t\t\t\t\t\t\t";
    static const char spaces[] = "    ";
    const int i = (indent > 15) ? 15 : indent;
    if ( i > 0 ) {
	if ( ! WriteRaw( tabs, i >> 1 ) ) {
	    return false;
	}
	if ( (i & 1) != 0 ) {
	    if ( ! WriteRaw( spaces, 4 ) ) {
		return false;
	    }
	}
    }

    // Write first string
    const unsigned s1_size = unsigned( StrLen( s1 ) );
    if ( ! WriteRaw( s1, s1_size ) ) {
	return false;
    }

    // Write alignment for the second string
    const unsigned s1_end = s1_size + (i << 2);
    if ( s1_end < 48 ) {
	unsigned pos = s1_end & (-8);
	while ( pos < 48 ) {
	    if ( ! WriteRaw( tabs, 1 ) ) {
		return false;
	    }
	    pos += 8;
	}
    }
    else {
	if ( ! WriteRaw( tabs, 1 ) ) {
	    return false;
	}
    }

    // Write second string
    if ( s2 != NULL ) {
	if ( ! WriteRaw( s2, unsigned(StrLen(s2)) ) ) {
	    return false;
	}
    }

    // Write EOL
    static const char eol[] = "\r\n";
    return WriteRaw( eol, 2 );
}

bool Output::Write( const char* s1, unsigned i )
{
    static char buf[256];
    wsprintf( buf, "%08X", i );
    return Write( s1, buf );
}

bool Output::Error( const char* s2 )
{
    return Write( "ERROR", s2 );
}

bool Output::CheckError( const char* fun, int error_code )
{
    if ( error_code < 0 ) {
	static char buf[64];
	wsprintf( buf, "%s failed with error %08X", fun, error_code );
	Error( buf );
	return false;
    }
    else {
	return true;
    }
}

Section::Section( Output& output, const char* section_name )
: output( output ), section_name( section_name )
{
    output.Write( "BEGIN", section_name );
    output.IncrementIndent();
}

Section::~Section()
{
    output.DecrementIndent();
    output.Write( "END", section_name );
}
